@@ -1,4 +1,4 @@
-# Copyright 2013-2014 DataStax, Inc.
+# Copyright 2013-2015 DataStax, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,29 +16,36 @@ from __future__ import absolute_import  # to enable import io from stdlib
 from collections import defaultdict, deque
 import errno
 from functools import wraps, partial
+from heapq import heappush, heappop
 import io
 import logging
-import os
+import six
+from six.moves import range
+import socket
+import struct
 import sys
-from threading import Event, RLock
+from threading import Thread, Event, RLock
 import time
 
+try:
+    import ssl
+except ImportError:
+    ssl = None  # NOQA
+
 if 'gevent.monkey' in sys.modules:
     from gevent.queue import Queue, Empty
 else:
     from six.moves.queue import Queue, Empty  # noqa
 
-import six
-from six.moves import range
-
 from cassandra import ConsistencyLevel, AuthenticationFailed, OperationTimedOut
-from cassandra.marshal import int32_pack, header_unpack, v3_header_unpack, int32_unpack
+from cassandra.marshal import int32_pack
 from cassandra.protocol import (ReadyMessage, AuthenticateMessage, OptionsMessage,
                                 StartupMessage, ErrorMessage, CredentialsMessage,
-                                QueryMessage, ResultMessage, decode_response,
+                                QueryMessage, ResultMessage, ProtocolHandler,
                                 InvalidRequestException, SupportedMessage,
                                 AuthResponseMessage, AuthChallengeMessage,
-                                AuthSuccessMessage, ProtocolException)
+                                AuthSuccessMessage, ProtocolException,
+                                MAX_SUPPORTED_VERSION, RegisterMessage)
 from cassandra.util import OrderedDict
 
 
@@ -88,9 +95,36 @@ HEADER_DIRECTION_FROM_CLIENT = 0x00
 HEADER_DIRECTION_TO_CLIENT = 0x80
 HEADER_DIRECTION_MASK = 0x80
 
-NONBLOCKING = (errno.EAGAIN, errno.EWOULDBLOCK)
+frame_header_v1_v2 = struct.Struct('>BbBi')
+frame_header_v3 = struct.Struct('>BhBi')
 
 
+class _Frame(object):
+    def __init__(self, version, flags, stream, opcode, body_offset, end_pos):
+        self.version = version
+        self.flags = flags
+        self.stream = stream
+        self.opcode = opcode
+        self.body_offset = body_offset
+        self.end_pos = end_pos
+
+    def __eq__(self, other):  # facilitates testing
+        if isinstance(other, _Frame):
+            return (self.version == other.version and
+                    self.flags == other.flags and
+                    self.stream == other.stream and
+                    self.opcode == other.opcode and
+                    self.body_offset == other.body_offset and
+                    self.end_pos == other.end_pos)
+        return NotImplemented
+
+    def __str__(self):
+        return "ver({0}); flags({1:04b}); stream({2}); op({3}); offset({4}); len({5})".format(self.version, self.flags, self.stream, self.opcode, self.body_offset, self.end_pos - self.body_offset)
+
+
+
+NONBLOCKING = (errno.EAGAIN, errno.EWOULDBLOCK)
+
 class ConnectionException(Exception):
     """
     An unrecoverable error was hit when attempting to use a connection,
@@ -109,6 +143,16 @@ class ConnectionShutdown(ConnectionException):
     pass
 
 
+class ProtocolVersionUnsupported(ConnectionException):
+    """
+    Server rejected startup message due to unsupported protocol version
+    """
+    def __init__(self, host, startup_version):
+        super(ProtocolVersionUnsupported, self).__init__("Unsupported protocol version on %s: %d",
+                                                         (host, startup_version))
+        self.startup_version = startup_version
+
+
 class ConnectionBusy(Exception):
     """
     An attempt was made to send a message through a :class:`.Connection` that
@@ -137,14 +181,22 @@ def defunct_on_error(f):
 
 DEFAULT_CQL_VERSION = '3.0.0'
 
+if six.PY3:
+    def int_from_buf_item(i):
+        return i
+else:
+    int_from_buf_item = ord
+
 
 class Connection(object):
 
+    CALLBACK_ERR_THREAD_THRESHOLD = 100
+
     in_buffer_size = 4096
     out_buffer_size = 4096
 
     cql_version = None
-    protocol_version = 2
+    protocol_version = MAX_SUPPORTED_VERSION
 
     keyspace = None
     compression = True
@@ -159,7 +211,7 @@ class Connection(object):
     in_flight = 0
 
     # A set of available request IDs.  When using the v3 protocol or higher,
-    # this will no initially include all request IDs in order to save memory,
+    # this will not initially include all request IDs in order to save memory,
     # but the set will grow if it is exhausted.
     request_ids = None
 
@@ -172,13 +224,25 @@ class Connection(object):
     lock = None
     user_type_map = None
 
+    msg_received = False
+
+    is_unsupported_proto_version = False
+
     is_control_connection = False
+    signaled_error = False  # used for flagging at the pool level
+
     _iobuf = None
+    _current_frame = None
+
+    _socket = None
+
+    _socket_impl = socket
+    _ssl_impl = ssl
 
     def __init__(self, host='127.0.0.1', port=9042, authenticator=None,
                  ssl_options=None, sockopts=None, compression=True,
-                 cql_version=None, protocol_version=2, is_control_connection=False,
-                 user_type_map=None):
+                 cql_version=None, protocol_version=MAX_SUPPORTED_VERSION, is_control_connection=False,
+                 user_type_map=None, connect_timeout=None):
         self.host = host
         self.port = port
         self.authenticator = authenticator
@@ -189,40 +253,27 @@ class Connection(object):
         self.protocol_version = protocol_version
         self.is_control_connection = is_control_connection
         self.user_type_map = user_type_map
+        self.connect_timeout = connect_timeout
         self._push_watchers = defaultdict(set)
+        self._requests = {}
         self._iobuf = io.BytesIO()
+
         if protocol_version >= 3:
-            self._header_unpack = v3_header_unpack
-            self._header_length = 5
             self.max_request_id = (2 ** 15) - 1
             # Don't fill the deque with 2**15 items right away. Start with 300 and add
             # more if needed.
             self.request_ids = deque(range(300))
             self.highest_request_id = 299
         else:
-            self._header_unpack = header_unpack
-            self._header_length = 4
             self.max_request_id = (2 ** 7) - 1
             self.request_ids = deque(range(self.max_request_id + 1))
             self.highest_request_id = self.max_request_id
 
-        # 0         8        16        24        32         40
-        # +---------+---------+---------+---------+---------+
-        # | version |  flags  |      stream       | opcode  |
-        # +---------+---------+---------+---------+---------+
-        # |                length                 |
-        # +---------+---------+---------+---------+
-        # |                                       |
-        # .            ...  body ...              .
-        # .                                       .
-        # .                                       .
-        # +----------------------------------------
-        self._full_header_length = self._header_length + 4
-
         self.lock = RLock()
+        self.connected_event = Event()
 
     @classmethod
-    def initialize_reactor(self):
+    def initialize_reactor(cls):
         """
         Called once by Cluster.connect().  This should be used by implementations
         to set up any resources that will be shared across connections.
@@ -230,13 +281,66 @@ class Connection(object):
         pass
 
     @classmethod
-    def handle_fork(self):
+    def handle_fork(cls):
         """
         Called after a forking.  This should cleanup any remaining reactor state
         from the parent process.
         """
         pass
 
+    @classmethod
+    def create_timer(cls, timeout, callback):
+        raise NotImplementedError()
+
+    @classmethod
+    def factory(cls, host, timeout, *args, **kwargs):
+        """
+        A factory function which returns connections which have
+        succeeded in connecting and are ready for service (or
+        raises an exception otherwise).
+        """
+        start = time.time()
+        kwargs['connect_timeout'] = timeout
+        conn = cls(host, *args, **kwargs)
+        elapsed = time.time() - start
+        conn.connected_event.wait(timeout - elapsed)
+        if conn.last_error:
+            if conn.is_unsupported_proto_version:
+                raise ProtocolVersionUnsupported(host, conn.protocol_version)
+            raise conn.last_error
+        elif not conn.connected_event.is_set():
+            conn.close()
+            raise OperationTimedOut("Timed out creating connection (%s seconds)" % timeout)
+        else:
+            return conn
+
+    def _connect_socket(self):
+        sockerr = None
+        addresses = socket.getaddrinfo(self.host, self.port, socket.AF_UNSPEC, socket.SOCK_STREAM)
+        for (af, socktype, proto, canonname, sockaddr) in addresses:
+            try:
+                self._socket = self._socket_impl.socket(af, socktype, proto)
+                if self.ssl_options:
+                    if not self._ssl_impl:
+                        raise Exception("This version of Python was not compiled with SSL support")
+                    self._socket = self._ssl_impl.wrap_socket(self._socket, **self.ssl_options)
+                self._socket.settimeout(self.connect_timeout)
+                self._socket.connect(sockaddr)
+                sockerr = None
+                break
+            except socket.error as err:
+                if self._socket:
+                    self._socket.close()
+                    self._socket = None
+                sockerr = err
+
+        if sockerr:
+            raise socket.error(sockerr.errno, "Tried connecting to %s. Last error: %s" % ([a[4] for a in addresses], sockerr.strerror or sockerr))
+
+        if self.sockopts:
+            for args in self.sockopts:
+                self._socket.setsockopt(*args)
+
     def close(self):
         raise NotImplementedError()
 
@@ -251,23 +355,49 @@ class Connection(object):
 
         self.last_error = exc
         self.close()
-        self.error_all_callbacks(exc)
+        self.error_all_requests(exc)
         self.connected_event.set()
         return exc
 
-    def error_all_callbacks(self, exc):
+    def error_all_requests(self, exc):
         with self.lock:
-            callbacks = self._callbacks
-            self._callbacks = {}
+            requests = self._requests
+            self._requests = {}
+
+        if not requests:
+            return
+
         new_exc = ConnectionShutdown(str(exc))
-        for cb in callbacks.values():
+        def try_callback(cb):
             try:
                 cb(new_exc)
             except Exception:
-                log.warning("Ignoring unhandled exception while erroring callbacks for a "
+                log.warning("Ignoring unhandled exception while erroring requests for a "
                             "failed connection (%s) to host %s:",
                             id(self), self.host, exc_info=True)
 
+        # run first callback from this thread to ensure pool state before leaving
+        cb, _ = requests.popitem()[1]
+        try_callback(cb)
+
+        if not requests:
+            return
+
+        # additional requests are optionally errored from a separate thread
+        # The default callback and retry logic is fairly expensive -- we don't
+        # want to tie up the event thread when there are many requests
+        def err_all_callbacks():
+            for cb, _ in requests.values():
+                try_callback(cb)
+        if len(requests) < Connection.CALLBACK_ERR_THREAD_THRESHOLD:
+            err_all_callbacks()
+        else:
+            # daemon thread here because we want to stay decoupled from the cluster TPE
+            # TODO: would it make sense to just have a driver-global TPE?
+            t = Thread(target=err_all_callbacks)
+            t.daemon = True
+            t.start()
+
     def get_request_id(self):
         """
         This must be called while self.lock is held.
@@ -288,14 +418,16 @@ class Connection(object):
             except Exception:
                 log.exception("Pushed event handler errored, ignoring:")
 
-    def send_msg(self, msg, request_id, cb):
+    def send_msg(self, msg, request_id, cb, encoder=ProtocolHandler.encode_message, decoder=ProtocolHandler.decode_message):
         if self.is_defunct:
             raise ConnectionShutdown("Connection to %s is defunct" % self.host)
         elif self.is_closed:
             raise ConnectionShutdown("Connection to %s is closed" % self.host)
 
-        self._callbacks[request_id] = cb
-        self.push(msg.to_binary(request_id, self.protocol_version, compression=self.compressor))
+        # queue the decoder function with the request
+        # this allows us to inject custom functions per request to encode, decode messages
+        self._requests[request_id] = (cb, decoder)
+        self.push(encoder(msg, request_id, self.protocol_version, compressor=self.compressor))
         return request_id
 
     def wait_for_response(self, msg, timeout=None):
@@ -348,86 +480,90 @@ class Connection(object):
             self.defunct(exc)
             raise
 
-    def register_watcher(self, event_type, callback):
-        raise NotImplementedError()
+    def register_watcher(self, event_type, callback, register_timeout=None):
+        """
+        Register a callback for a given event type.
+        """
+        self._push_watchers[event_type].add(callback)
+        self.wait_for_response(
+            RegisterMessage(event_list=[event_type]),
+            timeout=register_timeout)
 
-    def register_watchers(self, type_callback_dict):
-        raise NotImplementedError()
+    def register_watchers(self, type_callback_dict, register_timeout=None):
+        """
+        Register multiple callback/event type pairs, expressed as a dict.
+        """
+        for event_type, callback in type_callback_dict.items():
+            self._push_watchers[event_type].add(callback)
+        self.wait_for_response(
+            RegisterMessage(event_list=type_callback_dict.keys()),
+            timeout=register_timeout)
 
     def control_conn_disposed(self):
         self.is_control_connection = False
         self._push_watchers = {}
 
+    @defunct_on_error
+    def _read_frame_header(self):
+        buf = self._iobuf.getvalue()
+        pos = len(buf)
+        if pos:
+            version = int_from_buf_item(buf[0]) & PROTOCOL_VERSION_MASK
+            if version > MAX_SUPPORTED_VERSION:
+                raise ProtocolError("This version of the driver does not support protocol version %d" % version)
+            frame_header = frame_header_v3 if version >= 3 else frame_header_v1_v2
+            # this frame header struct is everything after the version byte
+            header_size = frame_header.size + 1
+            if pos >= header_size:
+                flags, stream, op, body_len = frame_header.unpack_from(buf, 1)
+                if body_len < 0:
+                    raise ProtocolError("Received negative body length: %r" % body_len)
+                self._current_frame = _Frame(version, flags, stream, op, header_size, body_len + header_size)
+        return pos
+
+    def _reset_frame(self):
+        self._iobuf = io.BytesIO(self._iobuf.read())
+        self._iobuf.seek(0, 2)  # io.SEEK_END == 2 (constant not present in 2.6)
+        self._current_frame = None
+
     def process_io_buffer(self):
         while True:
-            pos = self._iobuf.tell()
-            if pos < self._full_header_length or (self._total_reqd_bytes > 0 and pos < self._total_reqd_bytes):
+            if not self._current_frame:
+                pos = self._read_frame_header()
+            else:
+                pos = self._iobuf.tell()
+
+            if not self._current_frame or pos < self._current_frame.end_pos:
                 # we don't have a complete header yet or we
                 # already saw a header, but we don't have a
                 # complete message yet
                 return
             else:
-                # have enough for header, read body len from header
-                self._iobuf.seek(self._header_length)
-                body_len = int32_unpack(self._iobuf.read(4))
-
-                # seek to end to get length of current buffer
-                self._iobuf.seek(0, os.SEEK_END)
-                pos = self._iobuf.tell()
-
-                if pos >= body_len + self._full_header_length:
-                    # read message header and body
-                    self._iobuf.seek(0)
-                    msg = self._iobuf.read(self._full_header_length + body_len)
-
-                    # leave leftover in current buffer
-                    leftover = self._iobuf.read()
-                    self._iobuf = io.BytesIO()
-                    self._iobuf.write(leftover)
-
-                    self._total_reqd_bytes = 0
-                    self.process_msg(msg, body_len)
-                else:
-                    self._total_reqd_bytes = body_len + self._full_header_length
-                    return
+                frame = self._current_frame
+                self._iobuf.seek(frame.body_offset)
+                msg = self._iobuf.read(frame.end_pos - frame.body_offset)
+                self.process_msg(frame, msg)
+                self._reset_frame()
 
     @defunct_on_error
-    def process_msg(self, msg, body_len):
-        version, flags, stream_id, opcode = self._header_unpack(msg[:self._header_length])
+    def process_msg(self, header, body):
+        stream_id = header.stream
         if stream_id < 0:
             callback = None
+            decoder = ProtocolHandler.decode_message
         else:
-            callback = self._callbacks.pop(stream_id, None)
+            callback, decoder = self._requests.pop(stream_id, None)
             with self.lock:
                 self.request_ids.append(stream_id)
 
-        body = None
+        self.msg_received = True
+
         try:
-            # check that the protocol version is supported
-            given_version = version & PROTOCOL_VERSION_MASK
-            if given_version != self.protocol_version:
-                msg = "Server protocol version (%d) does not match the specified driver protocol version (%d). " +\
-                      "Consider setting Cluster.protocol_version to %d."
-                raise ProtocolError(msg % (given_version, self.protocol_version, given_version))
-
-            # check that the header direction is correct
-            if version & HEADER_DIRECTION_MASK != HEADER_DIRECTION_TO_CLIENT:
-                raise ProtocolError(
-                    "Header direction in response is incorrect; opcode %04x, stream id %r"
-                    % (opcode, stream_id))
-
-            if body_len > 0:
-                body = msg[self._full_header_length:]
-            elif body_len == 0:
-                body = six.binary_type()
-            else:
-                raise ProtocolError("Got negative body length: %r" % body_len)
-
-            response = decode_response(given_version, self.user_type_map, stream_id,
-                                       flags, opcode, body, self.decompressor)
+            response = decoder(header.version, self.user_type_map, stream_id,
+                               header.flags, header.opcode, body, self.decompressor)
         except Exception as exc:
             log.exception("Error decoding response from Cassandra. "
-                          "opcode: %04x; message contents: %r", opcode, msg)
+                          "%s; buffer: %r", header, self._iobuf.getvalue())
             if callback is not None:
                 callback(exc)
             self.defunct(exc)
@@ -436,6 +572,9 @@ class Connection(object):
         try:
             if stream_id >= 0:
                 if isinstance(response, ProtocolException):
+                    if 'unsupported protocol version' in response.message:
+                        self.is_unsupported_proto_version = True
+
                     log.error("Closing connection %s due to protocol error: %s", self, response.summary_msg())
                     self.defunct(response)
                 if callback is not None:
@@ -557,8 +696,8 @@ class Connection(object):
             else:
                 log.debug("Sending SASL-based auth response on %s", self)
                 initial_response = self.authenticator.initial_response()
-                initial_response = "" if initial_response is None else initial_response.encode('utf-8')
-                self.send_msg(AuthResponseMessage(initial_response), 0, self._handle_auth_response)
+                initial_response = "" if initial_response is None else initial_response
+                self.send_msg(AuthResponseMessage(initial_response), self.get_request_id(), self._handle_auth_response)
         elif isinstance(startup_response, ErrorMessage):
             log.debug("Received ErrorMessage on new connection (%s) from %s: %s",
                       id(self), self.host, startup_response.summary_msg())
@@ -673,6 +812,13 @@ class Connection(object):
 
         self.send_msg(query, request_id, process_result)
 
+    @property
+    def is_idle(self):
+        return not self.msg_received
+
+    def reset_idle(self):
+        self.msg_received = False
+
     def __str__(self):
         status = ""
         if self.is_defunct:
@@ -732,3 +878,182 @@ class ResponseWaiter(object):
             raise OperationTimedOut()
         else:
             return self.responses
+
+
+class HeartbeatFuture(object):
+    def __init__(self, connection, owner):
+        self._exception = None
+        self._event = Event()
+        self.connection = connection
+        self.owner = owner
+        log.debug("Sending options message heartbeat on idle connection (%s) %s",
+                  id(connection), connection.host)
+        with connection.lock:
+            if connection.in_flight < connection.max_request_id:
+                connection.in_flight += 1
+                connection.send_msg(OptionsMessage(), connection.get_request_id(), self._options_callback)
+            else:
+                self._exception = Exception("Failed to send heartbeat because connection 'in_flight' exceeds threshold")
+                self._event.set()
+
+    def wait(self, timeout):
+        self._event.wait(timeout)
+        if self._event.is_set():
+            if self._exception:
+                raise self._exception
+        else:
+            raise OperationTimedOut()
+
+    def _options_callback(self, response):
+        if not isinstance(response, SupportedMessage):
+            if isinstance(response, ConnectionException):
+                self._exception = response
+            else:
+                self._exception = ConnectionException("Received unexpected response to OptionsMessage: %s"
+                                                      % (response,))
+
+        log.debug("Received options response on connection (%s) from %s",
+                  id(self.connection), self.connection.host)
+        self._event.set()
+
+
+class ConnectionHeartbeat(Thread):
+
+    def __init__(self, interval_sec, get_connection_holders):
+        Thread.__init__(self, name="Connection heartbeat")
+        self._interval = interval_sec
+        self._get_connection_holders = get_connection_holders
+        self._shutdown_event = Event()
+        self.daemon = True
+        self.start()
+
+    class ShutdownException(Exception):
+        pass
+
+    def run(self):
+        self._shutdown_event.wait(self._interval)
+        while not self._shutdown_event.is_set():
+            start_time = time.time()
+
+            futures = []
+            failed_connections = []
+            try:
+                for connections, owner in [(o.get_connections(), o) for o in self._get_connection_holders()]:
+                    for connection in connections:
+                        self._raise_if_stopped()
+                        if not (connection.is_defunct or connection.is_closed):
+                            if connection.is_idle:
+                                try:
+                                    futures.append(HeartbeatFuture(connection, owner))
+                                except Exception:
+                                    log.warning("Failed sending heartbeat message on connection (%s) to %s",
+                                                id(connection), connection.host, exc_info=True)
+                                    failed_connections.append((connection, owner))
+                            else:
+                                connection.reset_idle()
+                        else:
+                            # make sure the owner sees this defunt/closed connection
+                            owner.return_connection(connection)
+                    self._raise_if_stopped()
+
+                for f in futures:
+                    self._raise_if_stopped()
+                    connection = f.connection
+                    try:
+                        f.wait(self._interval)
+                        # TODO: move this, along with connection locks in pool, down into Connection
+                        with connection.lock:
+                            connection.in_flight -= 1
+                        connection.reset_idle()
+                    except Exception:
+                        log.warning("Heartbeat failed for connection (%s) to %s",
+                                    id(connection), connection.host, exc_info=True)
+                        failed_connections.append((f.connection, f.owner))
+
+                for connection, owner in failed_connections:
+                    self._raise_if_stopped()
+                    connection.defunct(Exception('Connection heartbeat failure'))
+                    owner.return_connection(connection)
+            except self.ShutdownException:
+                pass
+            except Exception:
+                log.error("Failed connection heartbeat", exc_info=True)
+
+            elapsed = time.time() - start_time
+            self._shutdown_event.wait(max(self._interval - elapsed, 0.01))
+
+    def stop(self):
+        self._shutdown_event.set()
+        self.join()
+
+    def _raise_if_stopped(self):
+        if self._shutdown_event.is_set():
+            raise self.ShutdownException()
+
+
+class Timer(object):
+
+    canceled = False
+
+    def __init__(self, timeout, callback):
+        self.end = time.time() + timeout
+        self.callback = callback
+        if timeout < 0:
+            self.callback()
+
+    def cancel(self):
+        self.canceled = True
+
+    def finish(self, time_now):
+        if self.canceled:
+            return True
+
+        if time_now >= self.end:
+            self.callback()
+            return True
+
+        return False
+
+
+class TimerManager(object):
+
+    def __init__(self):
+        self._queue = []
+        self._new_timers = []
+
+    def add_timer(self, timer):
+        """
+        called from client thread with a Timer object
+        """
+        self._new_timers.append((timer.end, timer))
+
+    def service_timeouts(self):
+        """
+        run callbacks on all expired timers
+        Called from the event thread
+        :return: next end time, or None
+        """
+        queue = self._queue
+        if self._new_timers:
+            new_timers = self._new_timers
+            while new_timers:
+                heappush(queue, new_timers.pop())
+
+        if queue:
+            now = time.time()
+            while queue:
+                try:
+                    timer = queue[0][1]
+                    if timer.finish(now):
+                        heappop(queue)
+                    else:
+                        return timer.end
+                except Exception:
+                    log.exception("Exception while servicing timeout callback: ")
+
+    @property
+    def next_timeout(self):
+        try:
+            return self._queue[0][0]
+        except IndexError:
+            pass
