@@ -1,4 +1,4 @@
-# Copyright 2013-2014 DataStax, Inc.
+# Copyright 2013-2015 DataStax, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -24,16 +24,30 @@ import re
 import struct
 import time
 import six
+from six.moves import range, zip
 
 from cassandra import ConsistencyLevel, OperationTimedOut
-from cassandra.cqltypes import unix_time_from_uuid1
+from cassandra.util import unix_time_from_uuid1
 from cassandra.encoder import Encoder
 import cassandra.encoder
+from cassandra.protocol import _UNSET_VALUE
 from cassandra.util import OrderedDict
 
 import logging
 log = logging.getLogger(__name__)
 
+UNSET_VALUE = _UNSET_VALUE
+"""
+Specifies an unset value when binding a prepared statement.
+
+Unset values are ignored, allowing prepared statements to be used without specify
+
+See https://issues.apache.org/jira/browse/CASSANDRA-7304 for further details on semantics.
+
+.. versionadded:: 2.6.0
+
+Only valid when using native protocol v4+
+"""
 
 NON_ALPHA_REGEX = re.compile('[^a-zA-Z0-9]')
 START_BADCHAR_REGEX = re.compile('^[^a-zA-Z0-9]*')
@@ -103,11 +117,12 @@ def named_tuple_factory(colnames, rows):
     try:
         Row = namedtuple('Row', clean_column_names)
     except Exception:
-        log.warn("Failed creating named tuple for results with column names %s (cleaned: %s) (see Python 'namedtuple' documentation for details on name rules). "
-                 "Results will be returned with positional names. "
-                 "Avoid this by choosing different names, using SELECT \"<col name>\" AS aliases, "
-                 "or specifying a different row_factory on your Session" %
-                 (colnames, clean_column_names))
+        log.warning("Failed creating named tuple for results with column names %s (cleaned: %s) "
+                    "(see Python 'namedtuple' documentation for details on name rules). "
+                    "Results will be returned with positional names. "
+                    "Avoid this by choosing different names, using SELECT \"<col name>\" AS aliases, "
+                    "or specifying a different row_factory on your Session" %
+                    (colnames, clean_column_names))
         Row = namedtuple('Row', clean_column_names, rename=True)
 
     return [Row(*row) for row in rows]
@@ -166,6 +181,12 @@ class Statement(object):
     this will be set to a :class:`.QueryTrace` instance.
     """
 
+    trace_id = None
+    """
+    If :meth:`.Session.execute()` is run with `trace` set to :const:`True`,
+    this will be set to the tracing ID from the server.
+    """
+
     consistency_level = None
     """
     The :class:`.ConsistencyLevel` to be used for this operation.  Defaults
@@ -190,17 +211,27 @@ class Statement(object):
     :class:`~.TokenAwarePolicy` is configured for
     :attr:`.Cluster.load_balancing_policy`
 
-    It is set implicitly on :class:`.BoundStatement`, and :class:`.BatchStatement`, 
+    It is set implicitly on :class:`.BoundStatement`, and :class:`.BatchStatement`,
     but must be set explicitly on :class:`.SimpleStatement`.
 
     .. versionadded:: 2.1.3
     """
 
+    custom_payload = None
+    """
+    :ref:`custom_payload` to be passed to the server.
+
+    These are only allowed when using protocol version 4 or higher.
+
+    .. versionadded:: 2.6.0
+    """
+
     _serial_consistency_level = None
     _routing_key = None
 
     def __init__(self, retry_policy=None, consistency_level=None, routing_key=None,
-                 serial_consistency_level=None, fetch_size=FETCH_SIZE_UNSET, keyspace=None):
+                 serial_consistency_level=None, fetch_size=FETCH_SIZE_UNSET, keyspace=None,
+                 custom_payload=None):
         self.retry_policy = retry_policy
         if consistency_level is not None:
             self.consistency_level = consistency_level
@@ -211,6 +242,8 @@ class Statement(object):
             self.fetch_size = fetch_size
         if keyspace is not None:
             self.keyspace = keyspace
+        if custom_payload is not None:
+            self.custom_payload = custom_payload
 
     def _get_routing_key(self):
         return self._routing_key
@@ -247,6 +280,7 @@ class Statement(object):
             raise ValueError(
                 "serial_consistency_level must be either ConsistencyLevel.SERIAL "
                 "or ConsistencyLevel.LOCAL_SERIAL")
+        self._serial_consistency_level = serial_consistency_level
 
     def _del_serial_consistency_level(self):
         self._serial_consistency_level = None
@@ -289,8 +323,7 @@ class Statement(object):
 
 class SimpleStatement(Statement):
     """
-    A simple, un-prepared query.  All attributes of :class:`Statement` apply
-    to this class as well.
+    A simple, un-prepared query.
     """
 
     def __init__(self, query_string, *args, **kwargs):
@@ -298,6 +331,8 @@ class SimpleStatement(Statement):
         `query_string` should be a literal CQL statement with the exception
         of parameter placeholders that will be filled through the
         `parameters` argument of :meth:`.Session.execute()`.
+
+        All arguments to :class:`Statement` apply to this class as well
         """
         Statement.__init__(self, *args, **kwargs)
         self._query_string = query_string
@@ -326,9 +361,10 @@ class PreparedStatement(object):
     column_metadata = None
     query_id = None
     query_string = None
-    keyspace = None # change to prepared_keyspace in major release
+    keyspace = None  # change to prepared_keyspace in major release
 
     routing_key_indexes = None
+    _routing_key_index_set = None
 
     consistency_level = None
     serial_consistency_level = None
@@ -337,44 +373,43 @@ class PreparedStatement(object):
 
     fetch_size = FETCH_SIZE_UNSET
 
-    def __init__(self, column_metadata, query_id, routing_key_indexes, query, keyspace,
-                 protocol_version, consistency_level=None, serial_consistency_level=None,
-                 fetch_size=FETCH_SIZE_UNSET):
+    custom_payload = None
+
+    def __init__(self, column_metadata, query_id, routing_key_indexes, query,
+                 keyspace, protocol_version):
         self.column_metadata = column_metadata
         self.query_id = query_id
         self.routing_key_indexes = routing_key_indexes
         self.query_string = query
         self.keyspace = keyspace
         self.protocol_version = protocol_version
-        self.consistency_level = consistency_level
-        self.serial_consistency_level = serial_consistency_level
-        if fetch_size is not FETCH_SIZE_UNSET:
-            self.fetch_size = fetch_size
 
     @classmethod
-    def from_message(cls, query_id, column_metadata, cluster_metadata, query, prepared_keyspace, protocol_version):
+    def from_message(cls, query_id, column_metadata, pk_indexes, cluster_metadata, query, prepared_keyspace, protocol_version):
         if not column_metadata:
             return PreparedStatement(column_metadata, query_id, None, query, prepared_keyspace, protocol_version)
 
-        partition_key_columns = None
-        routing_key_indexes = None
+        if pk_indexes:
+            routing_key_indexes = pk_indexes
+        else:
+            routing_key_indexes = None
 
-        ks_name, table_name, _, _ = column_metadata[0]
-        ks_meta = cluster_metadata.keyspaces.get(ks_name)
-        if ks_meta:
-            table_meta = ks_meta.tables.get(table_name)
-            if table_meta:
-                partition_key_columns = table_meta.partition_key
+            first_col = column_metadata[0]
+            ks_meta = cluster_metadata.keyspaces.get(first_col.keyspace_name)
+            if ks_meta:
+                table_meta = ks_meta.tables.get(first_col.table_name)
+                if table_meta:
+                    partition_key_columns = table_meta.partition_key
 
-                # make a map of {column_name: index} for each column in the statement
-                statement_indexes = dict((c[2], i) for i, c in enumerate(column_metadata))
+                    # make a map of {column_name: index} for each column in the statement
+                    statement_indexes = dict((c.name, i) for i, c in enumerate(column_metadata))
 
-                # a list of which indexes in the statement correspond to partition key items
-                try:
-                    routing_key_indexes = [statement_indexes[c.name]
-                                           for c in partition_key_columns]
-                except KeyError:  # we're missing a partition key component in the prepared
-                    pass          # statement; just leave routing_key_indexes as None
+                    # a list of which indexes in the statement correspond to partition key items
+                    try:
+                        routing_key_indexes = [statement_indexes[c.name]
+                                               for c in partition_key_columns]
+                    except KeyError:  # we're missing a partition key component in the prepared
+                        pass          # statement; just leave routing_key_indexes as None
 
         return PreparedStatement(column_metadata, query_id, routing_key_indexes,
                                  query, prepared_keyspace, protocol_version)
@@ -382,11 +417,16 @@ class PreparedStatement(object):
     def bind(self, values):
         """
         Creates and returns a :class:`BoundStatement` instance using `values`.
-        The `values` parameter **must** be a sequence, such as a tuple or list,
-        even if there is only one value to bind.
+
+        See :meth:`BoundStatement.bind` for rules on input ``values``.
         """
         return BoundStatement(self).bind(values)
 
+    def is_routing_key_index(self, i):
+        if self._routing_key_index_set is None:
+            self._routing_key_index_set = set(self.routing_key_indexes) if self.routing_key_indexes else set()
+        return i in self._routing_key_index_set
+
     def __str__(self):
         consistency = ConsistencyLevel.value_to_name.get(self.consistency_level, 'Not Set')
         return (u'<PreparedStatement query="%s", consistency=%s>' %
@@ -398,8 +438,6 @@ class BoundStatement(Statement):
     """
     A prepared statement that has been bound to a particular set of values.
     These may be created directly or through :meth:`.PreparedStatement.bind()`.
-
-    All attributes of :class:`Statement` apply to this class as well.
     """
 
     prepared_statement = None
@@ -415,18 +453,20 @@ class BoundStatement(Statement):
     def __init__(self, prepared_statement, *args, **kwargs):
         """
         `prepared_statement` should be an instance of :class:`PreparedStatement`.
-        All other ``*args`` and ``**kwargs`` will be passed to :class:`.Statement`.
+
+        All arguments to :class:`Statement` apply to this class as well
         """
         self.prepared_statement = prepared_statement
 
         self.consistency_level = prepared_statement.consistency_level
         self.serial_consistency_level = prepared_statement.serial_consistency_level
         self.fetch_size = prepared_statement.fetch_size
+        self.custom_payload = prepared_statement.custom_payload
         self.values = []
 
         meta = prepared_statement.column_metadata
         if meta:
-            self.keyspace = meta[0][0]
+            self.keyspace = meta[0].keyspace_name
 
         Statement.__init__(self, *args, **kwargs)
 
@@ -434,82 +474,95 @@ class BoundStatement(Statement):
         """
         Binds a sequence of values for the prepared statement parameters
         and returns this instance.  Note that `values` *must* be:
+
         * a sequence, even if you are only binding one value, or
         * a dict that relates 1-to-1 between dict keys and columns
+
+        .. versionchanged:: 2.6.0
+
+            :data:`~.UNSET_VALUE` was introduced. These can be bound as positional parameters
+            in a sequence, or by name in a dict. Additionally, when using protocol v4+:
+
+            * short sequences will be extended to match bind parameters with UNSET_VALUE
+            * names may be omitted from a dict with UNSET_VALUE implied.
+
         """
         if values is None:
             values = ()
-        col_meta = self.prepared_statement.column_metadata
-
         proto_version = self.prepared_statement.protocol_version
+        col_meta = self.prepared_statement.column_metadata
+        col_meta_len = len(col_meta)
+        value_len = len(values)
 
         # special case for binding dicts
         if isinstance(values, dict):
-            dict_values = values
+            unbound_values = values.copy()
             values = []
 
             # sort values accordingly
             for col in col_meta:
                 try:
-                    values.append(dict_values[col[2]])
+                    values.append(unbound_values.pop(col.name))
                 except KeyError:
-                    raise KeyError(
-                        'Column name `%s` not found in bound dict.' %
-                        (col[2]))
+                    if proto_version >= 4:
+                        values.append(UNSET_VALUE)
+                    else:
+                        raise KeyError(
+                            'Column name `%s` not found in bound dict.' %
+                            (col.name))
 
-            # ensure a 1-to-1 dict keys to columns relationship
-            if len(dict_values) != len(col_meta):
-                # find expected columns
-                columns = set()
-                for col in col_meta:
-                    columns.add(col[2])
+            value_len = len(values)
 
-                # generate error message
-                if len(dict_values) > len(col_meta):
-                    difference = set(dict_values.keys()).difference(columns)
-                    msg = "Too many arguments provided to bind() (got %d, expected %d). " + \
-                          "Unexpected keys %s."
-                else:
-                    difference = set(columns).difference(dict_values.keys())
-                    msg = "Too few arguments provided to bind() (got %d, expected %d). " + \
-                          "Expected keys %s."
+            if unbound_values:
+                raise ValueError("Unexpected arguments provided to bind(): %s" % unbound_values.keys())
 
-                # exit with error message
-                msg = msg % (len(values), len(col_meta), difference)
-                raise ValueError(msg)
-
-        if len(values) > len(col_meta):
+        if value_len > col_meta_len:
             raise ValueError(
                 "Too many arguments provided to bind() (got %d, expected %d)" %
                 (len(values), len(col_meta)))
 
-        if self.prepared_statement.routing_key_indexes and \
-           len(values) < len(self.prepared_statement.routing_key_indexes):
+        # this is fail-fast for clarity pre-v4. When v4 can be assumed,
+        # the error will be better reported when UNSET_VALUE is implicitly added.
+        if proto_version < 4 and self.prepared_statement.routing_key_indexes and \
+           value_len < len(self.prepared_statement.routing_key_indexes):
             raise ValueError(
                 "Too few arguments provided to bind() (got %d, required %d for routing key)" %
-                (len(values), len(self.prepared_statement.routing_key_indexes)))
+                (value_len, len(self.prepared_statement.routing_key_indexes)))
 
         self.raw_values = values
         self.values = []
         for value, col_spec in zip(values, col_meta):
             if value is None:
                 self.values.append(None)
+            elif value is UNSET_VALUE:
+                if proto_version >= 4:
+                    self._append_unset_value()
+                else:
+                    raise ValueError("Attempt to bind UNSET_VALUE while using unsuitable protocol version (%d < 4)" % proto_version)
             else:
-                col_type = col_spec[-1]
-
                 try:
-                    self.values.append(col_type.serialize(value, proto_version))
-                except (TypeError, struct.error):
-                    col_name = col_spec[2]
-                    expected_type = col_type
+                    self.values.append(col_spec.type.serialize(value, proto_version))
+                except (TypeError, struct.error) as exc:
                     actual_type = type(value)
-
                     message = ('Received an argument of invalid type for column "%s". '
-                               'Expected: %s, Got: %s' % (col_name, expected_type, actual_type))
+                               'Expected: %s, Got: %s; (%s)' % (col_spec.name, col_spec.type, actual_type, exc))
                     raise TypeError(message)
 
+        if proto_version >= 4:
+            diff = col_meta_len - len(self.values)
+            if diff:
+                for _ in range(diff):
+                    self._append_unset_value()
+
         return self
 
+    def _append_unset_value(self):
+        next_index = len(self.values)
+        if self.prepared_statement.is_routing_key_index(next_index):
+            col_meta = self.prepared_statement.column_metadata[next_index]
+            raise ValueError("Cannot bind UNSET_VALUE as a part of the routing key '%s'" % col_meta.name)
+        self.values.append(UNSET_VALUE)
+
     @property
     def routing_key(self):
         if not self.prepared_statement.routing_key_indexes:
@@ -602,7 +655,8 @@ class BatchStatement(Statement):
     _session = None
 
     def __init__(self, batch_type=BatchType.LOGGED, retry_policy=None,
-                 consistency_level=None, serial_consistency_level=None, session=None):
+                 consistency_level=None, serial_consistency_level=None,
+                 session=None, custom_payload=None):
         """
         `batch_type` specifies The :class:`.BatchType` for the batch operation.
         Defaults to :attr:`.BatchType.LOGGED`.
@@ -613,6 +667,11 @@ class BatchStatement(Statement):
         `consistency_level` should be a :class:`~.ConsistencyLevel` value
         to be used for all operations in the batch.
 
+        `custom_payload` is a :ref:`custom_payload` passed to the server.
+        Note: as Statement objects are added to the batch, this map is
+        updated with any values found in their custom payloads. These are
+        only allowed when using protocol version 4 or higher.
+
         Example usage:
 
         .. code-block:: python
@@ -638,12 +697,15 @@ class BatchStatement(Statement):
 
         .. versionchanged:: 2.1.0
             Added `serial_consistency_level` as a parameter
+
+        .. versionchanged:: 2.6.0
+            Added `custom_payload` as a parameter
         """
         self.batch_type = batch_type
         self._statements_and_parameters = []
         self._session = session
         Statement.__init__(self, retry_policy=retry_policy, consistency_level=consistency_level,
-                           serial_consistency_level=serial_consistency_level)
+                           serial_consistency_level=serial_consistency_level, custom_payload=custom_payload)
 
     def add(self, statement, parameters=None):
         """
@@ -661,7 +723,7 @@ class BatchStatement(Statement):
         elif isinstance(statement, PreparedStatement):
             query_id = statement.query_id
             bound_statement = statement.bind(() if parameters is None else parameters)
-            self._maybe_set_routing_attributes(bound_statement)
+            self._update_state(bound_statement)
             self._statements_and_parameters.append(
                 (True, query_id, bound_statement.values))
         elif isinstance(statement, BoundStatement):
@@ -669,7 +731,7 @@ class BatchStatement(Statement):
                 raise ValueError(
                     "Parameters cannot be passed with a BoundStatement "
                     "to BatchStatement.add()")
-            self._maybe_set_routing_attributes(statement)
+            self._update_state(statement)
             self._statements_and_parameters.append(
                 (True, statement.prepared_statement.query_id, statement.values))
         else:
@@ -678,7 +740,7 @@ class BatchStatement(Statement):
             if parameters:
                 encoder = Encoder() if self._session is None else self._session.encoder
                 query_string = bind_params(query_string, parameters, encoder)
-            self._maybe_set_routing_attributes(statement)
+            self._update_state(statement)
             self._statements_and_parameters.append((False, query_string, ()))
         return self
 
@@ -697,6 +759,16 @@ class BatchStatement(Statement):
                 self.routing_key = statement.routing_key
                 self.keyspace = statement.keyspace
 
+    def _update_custom_payload(self, statement):
+        if statement.custom_payload:
+            if self.custom_payload is None:
+                self.custom_payload = {}
+            self.custom_payload.update(statement.custom_payload)
+
+    def _update_state(self, statement):
+        self._maybe_set_routing_attributes(statement)
+        self._update_custom_payload(statement)
+
     def __str__(self):
         consistency = ConsistencyLevel.value_to_name.get(self.consistency_level, 'Not Set')
         return (u'<BatchStatement type=%s, statements=%d, consistency=%s>' %
@@ -757,6 +829,13 @@ class QueryTrace(object):
     A :class:`datetime.timedelta` measure of the duration of the query.
     """
 
+    client = None
+    """
+    The IP address of the client that issued this request
+
+    This is only available when using Cassandra 2.2+
+    """
+
     coordinator = None
     """
     The IP address of the host that acted as coordinator for this request.
@@ -825,6 +904,8 @@ class QueryTrace(object):
             self.started_at = session_row.started_at
             self.coordinator = session_row.coordinator
             self.parameters = session_row.parameters
+            # since C* 2.2
+            self.client = getattr(session_row, 'client', None)
 
             log.debug("Attempting to fetch trace events for trace ID: %s", self.trace_id)
             time_spent = time.time() - start
@@ -836,14 +917,14 @@ class QueryTrace(object):
             break
 
     def _execute(self, query, parameters, time_spent, max_wait):
+        timeout = (max_wait - time_spent) if max_wait is not None else None
+        future = self._session._create_response_future(query, parameters, trace=False, custom_payload=None, timeout=timeout)
         # in case the user switched the row factory, set it to namedtuple for this query
-        future = self._session._create_response_future(query, parameters, trace=False)
         future.row_factory = named_tuple_factory
         future.send_request()
 
-        timeout = (max_wait - time_spent) if max_wait is not None else None
         try:
-            return future.result(timeout=timeout)
+            return future.result()
         except OperationTimedOut:
             raise TraceUnavailable("Trace information was not available within %f seconds" % (max_wait,))
 
